import sys


PY3 = sys.version_info > (3,)
PY2 = not PY3

if PY2:
    import ConfigParser as configparser
    from UserDict import IterableUserDict
    from Queue import Empty
    from urlparse import urlparse
    from urllib2 import HTTPError
    from urllib import quote_plus
    from cStringIO import StringIO
else:
    import configparser  # noqa: F401
    from collections import UserDict as IterableUserDict  # noqa: F401
    from queue import Empty  # noqa: F401
    from urllib.parse import quote_plus, urlparse  # noqa: F401
    from urllib.error import HTTPError  # noqa: F401
    from io import StringIO  # noqa: F401



import logging

from bottle import default_app

from ..toolkit.app import SubCommand
from .app import app


logger = logging.getLogger(__name__)


@app.command
class Routes(SubCommand):
    """ Show HTTP routing table."""

    def main(self, args):
        for path, method, callback in sorted(iter_bottle_routes()):
            doc = callback.__doc__ or ''
            try:
                title = doc.splitlines()[0]
            except IndexError:
                title = ''
            print("    %6.6s %-64s  %s" % (method, path, title))

        return 0


def iter_bottle_routes(bottle=None):
    if bottle is None:
        bottle = default_app()
    for route in bottle.routes:
        if 'PROXY' == route.method:  # Gateway to a plugin.
            prefix = route.config['mountpoint.prefix'].rstrip('/')
            subapp = route.config['mountpoint.target']
            dynamic = '<:re:' in route.rule
            for rule, method, callback in iter_bottle_routes(subapp):
                # Show root for static PROXY only.
                if dynamic:
                    if '/' == rule:
                        continue
                elif '/' != rule:
                    continue
                yield prefix + ('' if '/' == rule else rule), method, callback
        else:
            yield route.rule, route.method, route.callback



import logging
from logging.config import dictConfig
from logging.handlers import SysLogHandler
import sys


class ColoredStreamHandler(logging.StreamHandler):

    _color_map = {  # Colors code from systemctl/journalctl
        logging.DEBUG: '38;5;245',
        logging.INFO: '1;39',
        logging.WARN: '38;5;185',
        logging.ERROR: '31',
        logging.CRITICAL: '1;91',
    }

    def format(self, record):
        lines = logging.StreamHandler.format(self, record)
        color = self._color_map.get(record.levelno, '39')
        lines = ''.join([
            '\033[0;%sm%s\033[0m' % (color, line)
            for line in lines.splitlines(True)
        ])
        return lines


class MultilineFormatter(logging.Formatter):
    def format(self, record):
        s = logging.Formatter.format(self, record)
        if '\n' in s:
            s = s.replace('\n', '\n\t')
        return s


class NullHandler(logging.Handler):
    def handle(self, record):
        pass

    def emit(self, record):
        pass

    def createLock(self):
        self.lock = None


class SystemdFormatter(MultilineFormatter):
    # cf. http://0pointer.de/blog/projects/journal-submit.html
    priority_map = {
        logging.NOTSET: 6,
        logging.DEBUG: 7,
        logging.INFO: 6,
        logging.WARNING: 4,
        logging.ERROR: 3,
        logging.CRITICAL: 2,
    }

    def format(self, record):
        s = MultilineFormatter.format(self, record)
        prefix = '<%d>' % self.priority_map[record.levelno]
        lines = [prefix + line for line in s.splitlines(True)]
        return ''.join(lines)


class LastnameFilter(logging.Filter):
    root, _, _ = __name__.partition('.')

    def filter(self, record):
        record.lastname = record.name
        if record.name.startswith(self.root + '.'):
            _, record.lastname = record.name.rsplit('.', 1)
        # Always log, we are just enriching records.
        return 1


HANDLERS = {
    'file': {
        '()': 'logging.FileHandler',
        'mode': 'a',
        'formatter': 'dated_syslog',
    },
    'syslog': {
        '()': 'logging.handlers.SysLogHandler',
        'formatter': 'syslog',
    },
    'stderr': {
        '()': __name__ + '.NullHandler',
    }
}


def setup_logging(**kw):
    logging_config = generate_logging_config(**kw)
    dictConfig(logging_config)


def configure_debug(logging_config, core, debug):
    # If --debug or DEBUG=1, apply DEBUG to all core loggers
    if debug in (True, '__debug__'):
        debug = core

    if hasattr(debug, 'split'):
        debug = filter(None, debug.split(','))

    # Now apply debug level.
    if debug:
        for loggername in debug:
            logger = logging_config['loggers'].setdefault(loggername, {})
            logger['level'] = 'DEBUG'


def generate_logging_config(
        level=None, destination=None, facility='local0',
        method='stderr', debug=None, systemd=False, **kw):
    # Our logging strategy is to always log on stderr and allow to log on
    # either syslog or a file, depending on user configuration.
    #
    # stderr mean either console when launching in a terminal, file when stderr
    # is piped or systemd output. Thus we adapt stderr to these cases depending
    # on systemd an isatty().

    core = LastnameFilter.root

    if level is None:
        level = 'DEBUG' if debug else 'INFO'

    if debug is None:
        debug = level == 'DEBUG'

    verbose = debug or level == 'DEBUG'

    facility = SysLogHandler.facility_names[facility]
    HANDLERS['syslog']['facility'] = facility
    HANDLERS['syslog']['address'] = destination
    HANDLERS['file']['filename'] = destination

    stderr_handler = 'logging.StreamHandler'
    if sys.stderr.isatty():
        stderr_handler = __name__ + '.ColoredStreamHandler'

    minimal_fmt = '%(levelname)s:  %(lastname)s: %(message)s'
    verbose_fmt = '%(asctime)s ' + core + '[%(process)d] ' + minimal_fmt
    syslog_fmt = (
        core + "[%(process)d] %(levelname)s:  %(lastname)s: %(message)s"
    )

    logging_config = {
        'version': 1,
        'disable_existing_loggers': False,
        'filters': {
            'lastname': {
                '()': __name__ + '.LastnameFilter',
            }
        },
        'formatters': {
            'console': {
                '()': __name__ + '.MultilineFormatter',
                'format': verbose_fmt if verbose else minimal_fmt,
                'datefmt': '%Y-%m-%d %H:%M:%S %Z',
            },
            'dated_syslog': {
                '()': __name__ + '.MultilineFormatter',
                'format': '%(asctime)s ' + syslog_fmt,
                'datefmt': '%Y-%m-%d %H:%M:%S %Z',
            },
            'syslog': {
                '()': __name__ + '.MultilineFormatter',
                'format': syslog_fmt,
            },
            'systemd': {
                '()': __name__ + '.SystemdFormatter',
                'format': '%(lastname)s: %(message)s',
            }
        },
        'handlers': {
            'configured': dict(
                HANDLERS[method],
                filters=['lastname'],
            ),
            'stderr': {
                '()': stderr_handler,
                'formatter': 'systemd' if systemd else 'console',
                'filters': ['lastname'],
            },
        },
        'root': {
            'level': 'INFO',
            # Avoid instanciate all handlers, especially syslog which open
            # /dev/log
            'handlers': ['stderr', 'configured'],
        },
        'loggers': {},
    }

    # Apply level to temboard loggers only
    logging_config['loggers'][core] = dict(level=level)

    configure_debug(logging_config, core, debug)

    return logging_config



from ..toolkit.app import SubCommand
from .app import app


@app.command
class Web(SubCommand):
    """Standalone web server

    For testing purpose only. Some feature won't work without combined task
    manager processes. See serve command for production.

    """

    def main(self, args):

        self.app.httpd.run()

        return 0



import logging
import ssl
from socket import error as SocketError
from wsgiref.simple_server import (
    make_server,
    ServerHandler,
    WSGIRequestHandler,
)

from bottle import debug, default_app

from .. import __version__
from ..errors import UserError
from ..toolkit.services import Service


logger = logging.getLogger(__name__)


class HTTPDService(Service):
    def setup(self):
        ServerHandler.server_software = 'temBoard-agent/%s' % __version__

        try:
            bottle = default_app()
            debug(self.app.debug)

            self.server = make_server(
                self.app.config.temboard.address,
                self.app.config.temboard.port,
                app=bottle,
                handler_class=CustomWSGIRequestHandler,
            )
        except SocketError as e:
            raise UserError("Failed to start HTTPS server: {}.".format(e))
        try:
            logger.debug(
                "Using SSL key %s.", self.app.config.temboard.ssl_key_file)
            logger.debug(
                "Using SSL certificate %s.",
                self.app.config.temboard.ssl_cert_file)
            self.server.socket = ssl.wrap_socket(
                self.server.socket,
                keyfile=self.app.config.temboard.ssl_key_file,
                certfile=self.app.config.temboard.ssl_cert_file,
                server_side=True,
            )
        except Exception as e:
            raise UserError("Failed to setup SSL: {}.".format(e))
        self.server.timeout = 1

    def serve1(self):
        self.server.handle_request()


class CustomWSGIRequestHandler(WSGIRequestHandler):
    def get_environ(self):
        env = super(CustomWSGIRequestHandler, self).get_environ()

        # Save raw PATH_INFO for signature computation.
        path, _, _ = self.path.partition('?')
        env['RAW_PATH_INFO'] = path

        return env



import logging
import time

from bottle import Bottle, default_app

from ...toolkit import taskmanager
from ...toolkit.configuration import OptionSpec
from ...toolkit.utils import utcnow

from . import db
from . import metrics


bottle = Bottle()
logger = logging.getLogger(__name__)
workers = taskmanager.WorkerSet()


@bottle.get('/')
def dashboard():
    return metrics.get_metrics_queue(default_app().temboard.config)


@bottle.get('/config')
def dashboard_config():
    app = default_app().temboard
    return dict(
        scheduler_interval=app.config.dashboard.scheduler_interval,
        history_length=app.config.dashboard.history_length,
    )


@bottle.get('/live')
def dashboard_live(pgpool):
    return metrics.get_metrics(default_app().temboard, pgpool)


@bottle.get('/history')
def dashboard_history():
    return metrics.get_history_metrics_queue(default_app().temboard.config)


@bottle.get('/buffers')
def dashboard_buffers(pgconn):
    return metrics.get_buffers(pgconn)


@bottle.get('/hitratio')
def dashboard_hitratio(pgconn):
    return metrics.get_hitratio(pgconn)


@bottle.get('/active_backends')
def dashboard_active_backends(pgconn):
    return metrics.get_active_backends(pgconn)


@bottle.get('/cpu')
def dashboard_cpu():
    return metrics.get_cpu_usage()


@bottle.get('/loadaverage')
def dashboard_loadaverage():
    return metrics.get_loadaverage()


@bottle.get('/memory')
def dashboard_memory():
    return metrics.get_memory_usage()


@bottle.get('/hostname')
def dashboard_hostname():
    return metrics.get_hostname(default_app().temboard.config)


@bottle.get('/databases')
def dashboard_databases(pgconn):
    return metrics.get_databases(pgconn)


@workers.register(pool_size=1)
def dashboard_collector_worker(app, pool=None):
    logger.info("Running dashboard collector.")

    data = metrics.get_metrics(app, pool)

    # We don't want to store notifications in the history.
    data.pop('notifications', None)
    logger.debug(data)

    db.add_metric(
        app.config.temboard.home,
        'dashboard.db',
        time.time(),
        data,
        app.config.dashboard.history_length
    )

    logger.debug("Done")


BATCH_DURATION = 5 * 60  # 5 minutes


@workers.register(pool_size=1)
def dashboard_collector_batch_worker(app):
    # Loop each configured interval in the batch duration.
    interval = app.config.dashboard.scheduler_interval
    pool = None
    start = utcnow()
    elapsed = 0
    while elapsed < BATCH_DURATION:
        if elapsed > 0:
            # Throttle interval after first run.
            time.sleep(interval)

        try:
            pool = pool or app.postgres.pool()
        except Exception as e:
            logger.error("Failed to connect to Postgres: %s", e)
        else:
            try:
                for attempt in pool.retry_connection():
                    with attempt:
                        dashboard_collector_worker(app, pool)
            except Exception as e:
                logger.error("Dashboard collector error: %s", e)

        elapsed = utcnow() - start
        elapsed = elapsed.total_seconds()


class DashboardPlugin:
    PG_MIN_VERSION = (90400, 9.4)
    s = 'dashboard'
    option_specs = [
        OptionSpec(s, 'scheduler_interval', default=2, validator=int),
        OptionSpec(s, 'history_length', default=150, validator=int),
    ]
    del s

    def __init__(self, app, **kw):
        self.app = app
        self.app.config.add_specs(self.option_specs)

    def bootstrap(self):
        db.bootstrap(self.app.config.temboard.home, 'dashboard.db')

    def load(self):
        default_app().mount('/dashboard', bottle)
        self.app.worker_pool.add(workers)
        workers.schedule(
            id='dashboard_collector_batch',
            redo_interval=BATCH_DURATION,
        )(dashboard_collector_batch_worker)
        self.app.scheduler.add(workers)

    def unload(self):
        self.app.scheduler.remove(workers)
        self.app.worker_pool.remove(workers)
        self.app.config.remove_specs(self.option_specs)



import os.path
from textwrap import dedent


class QueryFiler(dict):
    # A dict loading it's values from SQL files in a directory.

    def __init__(self, path):
        self.path = path
        super(QueryFiler, self).__init__()

    def load(self):
        fmt = dedent("""\
        -- From %s
        %s
        """)

        for path in self.iter_files():
            filename = os.path.basename(path)
            name, ext = os.path.splitext(filename)
            with open(path) as fo:
                sql = fo.read()
            self[name] = fmt % (path, sql)

    def iter_files(self):
        for filename in os.listdir(self.path):
            if not filename.endswith('.sql'):
                continue
            yield self.path + '/' + filename



"""
The inventory contains support functions to discover the features of the host
(ex: OS, CPU info, MemInfo)

Currently, only Linux is supported.
"""

import os
import logging
import pwd

from ...inventory import (
    SysInfo,
    PgInfo,
)


def host_info(discover):
    """Gather system information."""
    sinfo = SysInfo()

    s = discover['system']
    hostinfo = {
        "os": s['os'],
        "os_version": s['os_version'],
        "cpu_arch": s['arch'],
        "hostname": s['fqdn'],
    }
    hostinfo.update(sinfo.cpu_info())
    hostinfo['memory_size'] = s['memory']
    hostinfo['ip_addresses'] = sinfo.ip_addresses()
    hostinfo['filesystems'] = sinfo.file_systems()
    hostinfo['os_flavor'] = s['distribution']
    return hostinfo


def instance_info(pool, dbnames, discover):
    """Gather PostgreSQL instance information."""
    p = discover['postgres']
    conninfo = pool.postgres.pqvars()
    instance_info = {
        'hostname': discover['system']['fqdn'],
        'instance': p.get('cluster_name'),
        'local_name': p.get('cluster_name'),
        'available': True,
        'host': conninfo['host'],
        'port': conninfo['port'],
        'user': conninfo['user'],
        'database': conninfo['database'],
        'password': conninfo['password']
    }

    # Try the connection
    try:
        conn = pool.getconn(dbname=conninfo['database'])
    except Exception as e:
        logging.error("Failed to connect to Postgres: %s", e)
        instance_info['available'] = False
        return instance_info

    # Get PostgreSQL informations using PgInfo
    pginfo = PgInfo(conn)
    # Gather the info while where are connected
    instance_info['version_num'] = p['version_num']
    instance_info['version'] = p['version']
    instance_info['data_directory'] = p['data_directory']
    instance_info['max_connections'] = p['max_connections']
    instance_info['start_time'] = pginfo.start_time()

    # hot standby is available from 9.0
    instance_info['standby'] = pginfo.is_in_recovery()

    # Grab the list of tablespaces
    instance_info['tablespaces'] = pginfo.tablespaces(
        instance_info['data_directory'])

    # When the user has not given a dbnames list or '*' in the
    # configuration file, we must get the list of databases. Since
    # we have a working connection, let's do it now.
    dbs = pginfo.databases()
    instance_info['dbnames'] = []
    for db in dbnames:
        if db == '*':
            instance_info['dbnames'] = list(dbs.values())
            break
        if db in dbs.keys():
            instance_info['dbnames'].append(dbs[db])

    # Now that we have the data_directory, find the owner
    try:
        statinfo = os.stat(instance_info['data_directory'])
        instance_info['sysuser'] = pwd.getpwuid(statinfo.st_uid).pw_name
    except OSError as e:
        logging.warning("Unable to get the owner of PGDATA: %s", str(e))
        instance_info['sysuser'] = None

    return instance_info



# coding: utf-8

import bdb
import pkg_resources
from distutils.util import strtobool
from glob import glob
import logging
import os
import pdb
from codecs import open
from site import main as refresh_pythonpath
from textwrap import dedent
import sys
from argparse import (
    ArgumentParser,
    Action as ArgAction,
    SUPPRESS as SUPPRESS_ARG,
    RawDescriptionHelpFormatter,
)

try:
    from pkg_resources import (
        _initialize_master_working_set as refresh_distributions
    )
except ImportError:  # pragma: nocover
    def refresh_distributions():
        logger.info("setuptools is too old for plugin hotload.")
        logger.info("If new plugin can't be found, just restart.")

from .configuration import MergedConfiguration, OptionSpec
from .log import setup_logging, LastnameFilter
from .errors import UserError
from . import validators as v
from .pycompat import configparser


logger = logging.getLogger(__name__)


class StoreDefinedAction(ArgAction):
    def __call__(self, parser, namespace, values, option_string=None):
        # Store True if argument is defined.
        if values is None:
            values = True
        setattr(namespace, self.dest, values)


class BaseApplication(object):
    # This object contains application context and logic.
    #
    # The core logic is managing configuration and plugins, this is the
    # bootstrap process. Once the app is ready, it owns objects representing
    # the state of the app : config, plugins, etc. Each script or plugin can
    # add an object, it will be shared with other.

    PROGRAM = "temboard"
    VERSION = "unknown"
    REPORT_URL = "https://github.com/dalibo/temboard/issues/new"

    DEFAULT_CONFIGFILES = [
        '%s.conf' % (LastnameFilter.root),
        '/etc/%s/%s.conf' % (LastnameFilter.root, LastnameFilter.root),
    ]
    DEFAULT_PLUGINS_EP = LastnameFilter.root + '.plugins'
    DEFAULT_PLUGINS = []

    def __init__(self, specs=None, with_plugins=DEFAULT_PLUGINS_EP, main=None):
        # If `None`, plugin loading is disabled.
        self.with_plugins = with_plugins
        self.plugins = {}
        self.config = MergedConfiguration()
        # This dict stores env, args and parser for hot reloading of
        # configuration.
        self.config_sources = dict()
        self.config_specs = self.init_specs(specs)
        # Active options specs for multi-stage parsing.
        self.active_config_specs = []
        self.services = []
        self._main = main
        self.commands = {}

    def __repr__(self):
        return '<%s>' % (self.__class__.__name__)

    def bootstrap(self, args, environ, service=False):
        # bootstrapping the app is a complex process to manage options loading
        # incrementally.

        # Whether we are oneshot CLI or long running service.
        self.is_service = service
        config = self.config
        # Stage 1: Read configfile option
        config.add_specs(self.list_stage1_specs())
        self.config_sources.update(dict(args=args, environ=environ))
        config.load(**self.config_sources)

        # Stage 2: Now read configfile
        parser = configparser.RawConfigParser()
        configfile = self.find_config_file()
        if configfile is None:
            logger.info("No config file found.")
        else:
            logger.info("Using config file %s.", configfile)
            self.config.temboard['configfile'] = configfile
            self.read_file(parser, configfile)
            self.read_dir(parser, configfile + '.d')
            self.config_sources.update(dict(
                parser=parser, pwd=os.path.dirname(configfile),
            ))

        # Stage 3: Add core and app specific options and load them.
        config.add_specs(self.config_specs.values())
        config.load(**self.config_sources)

        # Save loaded file.
        self.config['temboard']['configfile'] = configfile

        return self.config

    def init_specs(self, app_specs):
        # Declare all option specs.

        specs = dict()

        def add_specs(*new_specs):
            for spec in new_specs:
                specs[str(spec)] = spec

        s = 'temboard'
        add_specs(OptionSpec(
            s, 'configfile',
            validator=v.file_,
        ))
        if self.with_plugins:
            add_specs(OptionSpec(
                s, 'plugins', default=self.DEFAULT_PLUGINS,
                validator=v.jsonlist,
            ))

        s = 'logging'
        add_specs(
            OptionSpec(s, 'debug', default=False),
            OptionSpec(s, 'method', default='stderr', validator=v.logmethod),
            OptionSpec(s, 'level', default='INFO', validator=v.loglevel),
            OptionSpec(
                s, 'facility', default='local0', validator=v.syslogfacility),
            OptionSpec(s, 'destination', default='/dev/log'),
        )

        if app_specs:
            add_specs(*app_specs)

        return specs

    def list_stage1_specs(self):
        # List options specs required for bootstrap from args and environ:
        # configfile.
        return [self.config_specs[name] for name in [
            'temboard_configfile',
            # Allow to enable debug as soon as possible. Other options will
            # keep defaults.
            'logging_debug',
        ]]

    def create_parser(self, *a, **kw):
        kw.setdefault('argument_default', SUPPRESS_ARG)
        _, d = extract_help_description_from_docstring(self.__class__.__doc__)
        kw.setdefault('description', d)
        kw.setdefault('prog', self.PROGRAM)
        kw.setdefault('formatter_class', RawDescriptionHelpFormatter)
        return ArgumentParser(*a, **kw)

    def command(self, cls):
        # Class-decorator to instanciate and register a subcommand.
        command = cls.singleton = cls(self)
        self.commands[command.fullname] = command
        return cls

    def define_arguments(self, parser):
        # Configure an argparse parser for each subcommands declared with
        # @self.command class decorator.

        if not self.commands:
            return

        subparsers = parser.add_subparsers(
            title="Available commands",
            metavar="COMMAND",
            help="Name of one sub-command described bellow.",
        )

        for fullname, command in self.commands.items():
            if "." in fullname:
                continue  # Let commands declare their sub-commands.

            h, d = extract_help_description_from_docstring(command.__doc__)
            subparser = subparsers.add_parser(
                command.name,
                help=h, description=d,
                formatter_class=parser.formatter_class,
                argument_default=parser.argument_default,
            )
            subparser.set_defaults(command_fullname=fullname)
            command.define_arguments(subparser)

    def apply_config(self):
        # Once config is loaded or reloaded, update application state to match
        # new configuration.

        try:
            self.setup_logging()
        except Exception as e:
            raise UserError("Failed to setup logging: %s." % (e,))
        for service in self.services:
            service.apply_config()

        if not self.with_plugins:
            return

        old_plugins = self.purge_plugins()
        new_plugins = self.create_plugins()
        if new_plugins:
            logger.debug("Reading new plugins configuration.")
            self.config.load(**self.config_sources)
        self.update_plugins(old_plugins=old_plugins)

    def find_config_file(self):
        configfile = self.config.temboard.configfile
        if configfile is None:
            for configfile in self.DEFAULT_CONFIGFILES:
                configfile = os.path.abspath(configfile)
                if os.path.exists(configfile):
                    break
            else:
                configfile = None
        return configfile

    def read_file(self, parser, filename):
        logger.debug('Reading %s.', filename)
        try:
            with open(filename, 'r', 'utf-8') as fp:
                parser.readfp(fp)
        except IOError as e:
            raise UserError(str(e))

    def read_dir(self, parser, dirname):
        if not os.path.isdir(dirname):
            return
        for filename in sorted(glob(dirname + '/*.conf')):
            self.read_file(parser, filename)

    def fetch_plugin(self, name):
        logger.debug("Looking for plugin %s.", name)
        for ep in pkg_resources.iter_entry_points(self.with_plugins, name):
            logger.debug("Found plugin %s.", ep)
            try:
                return ep.load()
            except Exception:
                logger.exception("Error while loading %s.", ep)
                raise UserError("Failed to load %s." % (ep.name,))
        else:
            raise UserError("Missing plugin: %s." % (name,))

    def create_plugins(self):
        self.config.plugins = dict()

        # Filter legacy plugins
        ng_plugins = filter(
            lambda name: name not in self.config.plugins,
            self.config.temboard.plugins
        )
        # Filter already loaded plugins
        unloaded_names = [
            n for n in ng_plugins
            if n not in self.plugins
        ]

        # Refresh sys.path and working_set to ensure new code is loadable.
        refresh_pythonpath()
        refresh_distributions()

        for name in unloaded_names:
            cls = self.fetch_plugin(name)
            plugin = cls(self)
            self.plugins[name] = plugin
            self.config.plugins.pop(name, None)

        return unloaded_names

    def update_plugins(self, old_plugins=None):
        # Load and unload plugins
        old_names = set(old_plugins or [])
        new_names = set(self.plugins)

        to_unload = old_names - new_names
        for name in to_unload:
            logger.info("Unloading plugin %s.", name)
            old_plugins[name].unload()

        to_load = new_names - old_names
        for name in to_load:
            logger.debug("Loading plugin %s.", name)
            self.plugins[name].load()

    def purge_plugins(self):
        old_plugins = self.plugins.copy()
        for name in list(self.plugins):
            if name in self.config.temboard.plugins:
                continue
            del self.plugins[name]
        return old_plugins

    def reload(self):
        logger.warning("Reloading configuration.")

        # Reset file parser and load values.
        self.config_sources['parser'] = parser = configparser.RawConfigParser()
        self.read_file(parser, self.config.temboard.configfile)
        self.read_dir(parser, self.config.temboard.configfile + '.d')
        self.config.load(reload_=True, **self.config_sources)

        self.apply_config()
        logger.info("Configuration reloaded.")
        return self

    def setup_logging(self):
        if self.config.logging.method != 'stderr' and not self.is_service:
            # Enforce stderr method for one shot command, this avoid creating a
            # logfile with bad privileges, and spam syslog or logfile with
            # command logs.
            logger.debug(
                "Disabling log method %s for one shot command.",
                self.config.logging.method,
            )
            self.config.logging.method = 'stderr'

        setup_logging(
            systemd='SYSTEMD' in os.environ,
            **self.config.logging)

    def __call__(self, argv=sys.argv[1:], environ=os.environ):
        return self.entrypoint(argv, environ)

    def entrypoint(self, argv, environ):
        self.debug = detect_debug_mode(environ)

        retcode = 1
        try:
            setup_logging(debug=self.debug)
            logger.debug("Starting %s %s.", self.PROGRAM, self.VERSION)
            retcode = self.main(argv, environ)
        except KeyboardInterrupt:
            logger.info('Terminated.')
        except bdb.BdbQuit:
            logger.info("Graceful exit from debugger.")
        except UserError as e:
            retcode = e.retcode
            logger.critical("%s", e)
        except Exception:
            logger.exception('Unhandled error:')
            if self.debug:
                pdb.post_mortem(sys.exc_info()[2])
            else:
                logger.error(
                    "%s version is %s.", LastnameFilter.root, self.VERSION)
                logger.error("This is a bug!")
                logger.error(
                    "Please report traceback to %s! Thanks!",
                    self.REPORT_URL,
                )
        exit(retcode)

    def main(self, argv, environ):
        if self._main is None:
            raise NotImplementedError()
        else:
            return self._main(self, argv, environ)


class SubCommand(object):
    # Base class for sub-command.
    #
    # Almost everything keeps in app object. Sub commands is roughly a set of
    # two fuctions : one to define argparse arguments, the other is the main
    # code to execute after app initialization.
    #
    # Class docstring is injected in argparse help for the command.

    name = None
    # Whether oneshot CLI or long running service. This is handled by
    # Application.setup_logging.
    is_service = False

    def __init__(self, parent):
        self.parent = parent  # The app or another command.

        names = []
        root = self.parent
        while hasattr(root, 'parent'):
            names[0:] = [root.name]
            root = root.parent

        if names:
            names.append("")    # Add final .

        if not self.name:
            self.name = self.__class__.__name__.lower()
        self.app = root
        self.prefix = ".".join(names)
        self.fullname = self.prefix + self.name

    def __repr__(self):
        return '<%s>' % (self.__class__.__name__)

    @classmethod
    def command(cls, subcommand_cls):
        # Class decorator to instanciate and register a subcommand.
        self = cls.singleton
        command = subcommand_cls.singleton = subcommand_cls(self)
        self.app.commands[command.fullname] = command
        return subcommand_cls

    @property
    def commands(self):
        app = self.app
        prefix = self.fullname + '.'
        my_commands = {}
        for fullname, command in app.commands.items():
            if not fullname.startswith(prefix):
                continue
            my_commands[fullname] = command

        return my_commands

    def define_arguments(self, parser):
        # Configure an argparse parser for each subcommands declared with
        # @self.command class decorator.

        my_commands = self.commands
        if not my_commands:
            return

        subparsers = parser.add_subparsers(
            title="Available commands",
            metavar="COMMAND",
            help="Name of one sub-command described bellow.",
        )

        for fullname, command in my_commands.items():
            subparser = subparsers.add_parser(
                command.name, help=command.__doc__)
            subparser.set_defaults(command_fullname=fullname)
            command.define_arguments(subparser)

    def main(self, args):
        raise NotImplementedError()


def detect_debug_mode(environ):
    debug = environ.get('DEBUG', '0')
    try:
        debug = bool(strtobool(debug))
        if debug:
            environ['TEMBOARD_LOGGING_DEBUG'] = '__debug__'
    except ValueError:
        environ['TEMBOARD_LOGGING_DEBUG'] = str(debug)
    return debug


def define_core_arguments(parser, appversion=None):
    if appversion:
        parser.add_argument(
            '-V', '--version',
            action='version',
            version=appversion,
        )
    parser.add_argument(
        '-c', '--config',
        action='store', dest='temboard_configfile',
        help="Configuration file", metavar='CONFIGFILE',
    )
    parser.add_argument(
        '--debug',
        action='store_const', const='temboardui', dest='logging_debug',
        help="Enable verbose messages for temBoard.",
    )


def extract_help_description_from_docstring(docstring):
    docstring = docstring or ""

    if "\n " in docstring:
        title, description = docstring.split("\n", 1)
        description = "%s\n%s" % (title, dedent(description))
    else:
        title = description = docstring

    return title, description



import logging

from bottle import Bottle, default_app, request

from . import functions as activity_functions
from ...notification import NotificationMgmt, Notification
from ...tools import validate_parameters


bottle = Bottle()
logger = logging.getLogger(__name__)


@bottle.get('/')
def get_activity(pgconn):
    return activity_functions.get_activity(pgconn)


@bottle.get('/waiting')
def get_activity_waiting(pgconn):
    return activity_functions.get_activity_waiting(pgconn)


@bottle.get('/blocking')
def get_activity_blocking(pgconn):
    return activity_functions.get_activity_blocking(pgconn)


@bottle.post('/kill')
def post_activity_kill(pgconn):
    app = default_app().temboard
    validate_parameters(request.json, [
        ('pids', int, True)
    ])
    backends = []
    for pid in request.json['pids']:
        killed = pgconn.queryscalar(
            "SELECT pg_terminate_backend(%s) AS killed" % (pid))
        backends.append(dict(pid=pid, killed=killed))

        # Push a notification.
        try:
            NotificationMgmt.push(
                app.config,
                Notification(
                    username=request.username,
                    message="Backend %s terminated" % (pid)
                )
            )
        except Exception:
            logger.exception("Failed to push notification:")

    return dict(backends=backends)


class ActivityPlugin:
    PG_MIN_VERSION = (90400, 9.4)

    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        default_app().mount('/activity', bottle)



import errno
import functools
import sys
import time
import uuid
import logging
import os.path
import signal
from ast import literal_eval
from select import select, error as SelectError
from datetime import datetime
from collections import deque
from multiprocessing import AuthenticationError, Process, Queue
from multiprocessing.connection import Listener, Client
from textwrap import dedent

from .services import Service
from .errors import StorageEngineError, UserError
from .perf import PerfCounters
from .pycompat import PY2, Empty


TM_DEF_LISTENER_ADDR = '/tmp/.temboardsched.sock'

# Message types
MSG_TYPE_TASK_NEW = 0
MSG_TYPE_TASK_STATUS = 1
MSG_TYPE_TASK_CANCEL = 2
MSG_TYPE_TASK_ABORT = 3
MSG_TYPE_TASK_LIST = 4
MSG_TYPE_RESP = 5
MSG_TYPE_ERROR = 6
MSG_TYPE_CONTEXT = 7

# Task status
TASK_STATUS_DEFAULT = 1
TASK_STATUS_SCHEDULED = 2
TASK_STATUS_QUEUED = 4
TASK_STATUS_DOING = 8
TASK_STATUS_DONE = 16
TASK_STATUS_FAILED = 32
TASK_STATUS_CANCELED = 64
TASK_STATUS_ABORTED = 128
TASK_STATUS_ABORT = 256

logger = logging.getLogger(__name__)


def ensure_str(value):
    # This code is used to instanciate multiprocessing.connection.Client. It
    # requires a str object in both Python 2 and 3.
    if type(value) is not str:
        if PY2:
            # From unicode to str.
            value = value.encode('utf-8')
        else:
            # From bytes to str.
            value = value.decode('utf-8')
    return value


def make_worker_definition(function, pool_size):
    return {
        'name': function.__name__,
        'pool_size': pool_size,
        'module': function.__module__,
        'function': function.__name__
    }


class Task(object):

    def __init__(self, worker_name=None, options=None, id=None,
                 status=TASK_STATUS_DEFAULT, start_datetime=None,
                 redo_interval=None, stop_datetime=None, output=None,
                 expire=3600):
        self.worker_name = worker_name
        self.options = options
        self.status = status
        self.start_datetime = start_datetime or datetime.utcnow()
        self.redo_interval = redo_interval
        self.stop_datetime = stop_datetime
        self.id = id
        self.output = output
        # Task expiration timeout in seconds.
        # Ended Tasks are removed from memory when current time exceeds
        # self.stop_datetime + self.expire.
        self.expire = expire

    def __repr__(self):
        return str(self.__dict__)


class Message(object):

    def __init__(self, type, content):
        self.type = type,
        self.content = content

    def __repr__(self):
        return str(self.__dict__)


class TaskList(object):

    def __init__(self, engine):
        self.engine = engine

    def recover(self):
        self.engine.recover(
            st_doing=(TASK_STATUS_DOING | TASK_STATUS_QUEUED),
            st_aborted=TASK_STATUS_ABORTED,
            st_scheduled=TASK_STATUS_SCHEDULED,
            st_default=TASK_STATUS_DEFAULT,
            now=datetime.utcnow()
        )

    def push(self, task):
        # Add a new task to the list
        if not task.id:
            task.id = self._gen_task_id()
        # Insert the task
        self.engine.insert(task)
        return task.id

    def get(self, id):
        return self.engine.get(id)

    def update(self, id, **kwargs):
        task = self.engine.get(id)
        if not task:
            raise Exception("Task id=%s not found" % id)

        for k, v in kwargs.items():
            try:
                getattr(task, k)
            except AttributeError:
                raise Exception("Task attribute %s does not exist" % k)
            setattr(task, k, v)
        self.engine.update(task)

    def rm(self, id):
        self.engine.delete(id)

    def _gen_task_id(self):
        id = str(uuid.uuid4())[0:8]

        while self.engine.exists(id):
            # If a task with the same id exists, try with a new random id
            id = str(uuid.uuid4())[0:8]

        return id

    def get_n_todo(self):
        # Return the number of ongoing tasks (QUEUED | DOING)
        return self.engine.count_by_status(
            TASK_STATUS_QUEUED | TASK_STATUS_DOING
        )

    def list(self):
        return self.engine.list()

    def list_to_do(self, status, now, redo=False):
        return self.engine.list_to_do(status, now, redo)

    def purge(self, status, now):
        return self.engine.purge(status, now)


class TaskManager(object):
    @staticmethod
    def send_message(address, message, authkey=''):
        conn = Client(ensure_str(address), authkey=authkey)
        conn.send(message)
        res = conn.recv()
        conn.close()
        return res


class Scheduler(object):
    trace = False

    def __init__(self, address, authkey):
        # Listener for TM -> Scheduler IPC
        self.listener = None
        # Listener address
        self.address = address
        # Listener authentication key
        self.authkey = authkey
        # Queue used to send Task orders (start, stop) to WorkerPool
        self.task_queue = None
        # Queue used to notify Scheduler about Task status
        self.event_queue = None
        self.task_list = None
        self.context = dict()
        self.worker_pool = None
        self.shutdown = None
        self.task_list_engine = None
        self.last_schedule = 0
        self.last_vacuum = 0
        self.select_timeout = None

    def set_context(self, key, val):
        self.context[key] = val

    def get_context(self):
        return self.context

    def handle_listener_message(self):
        # read message from the listener
        try:
            conn = self.listener.accept()
        except AuthenticationError as e:
            logger.exception(e)
            logger.error("Authentication failed")
            return

        try:
            message = conn.recv()
        except Exception as e:
            conn.close()
            logger.exception(e)
            logger.error("Unable to read the message")
            return

        message.type = message.type[0]
        if self.trace:
            logger.debug("Received Message=%s" % message)

        # handle incoming message and return a response
        conn.send(self.handle_message(message))
        conn.close()

    def handle_event_queue_message(self):
        # read message from the event queue
        try:
            message = self.event_queue.get(False)
        except Empty:
            # should not happen
            logger.error("Event queue empty")
            return

        message.type = message.type[0]
        if self.trace:
            logger.debug("Received Message=%s" % message)

        # handle incoming message
        self.handle_message(message)

    def setup(self):
        # Need to shutdown ?
        self.shutdown = False
        # Instanciate a new Listener
        self.listener = Listener(self.address, family='AF_UNIX',
                                 authkey=self.authkey)
        # TODO
        # self.sync_bootstrap_options()
        self.select_timeout = 1

    def serve1(self):
        # wait for I/O on Listener and event Queue
        try:
            fds, _, _ = select(
                [self.listener._listener._socket.fileno(),
                 self.event_queue._reader.fileno()],
                [],
                [],
                self.select_timeout
            )
        except SelectError as e:
            errno_, message = e.args
            if errno_ == errno.EINTR:
                # Interrupted by e.g. SIGHUP. Just stop.
                return
            else:
                raise

        if len(fds) > 0:
            for fd in fds:
                if fd == self.listener._listener._socket.fileno():
                    if self.shutdown:
                        # during shutdown we don't handle incoming
                        # message anymore
                        continue
                    self.handle_listener_message()
                elif fd == self.event_queue._reader.fileno():
                    self.handle_event_queue_message()

        if self.last_schedule < (time.time() - 1):
            self.schedule()
            self.last_schedule = time.time()

        if self.last_vacuum < (time.time() - 3600):
            self.task_list.engine.vacuum()
            self.last_vacuum = time.time()

    def setup_task_list(self):
        # Instanciate TaskList
        self.task_list = TaskList(self.task_list_engine)
        # Bootstrap storage
        self.task_list.engine.bootstrap()
        # Reset task status
        self.task_list.recover()

    def schedule(self):
        now = datetime.utcnow()

        self.task_list.purge(
            (TASK_STATUS_DONE | TASK_STATUS_FAILED | TASK_STATUS_ABORTED |
             TASK_STATUS_CANCELED),
            now
        )

        if self.shutdown:
            return

        to_do = self.task_list.list_to_do(TASK_STATUS_DEFAULT, now)

        for task in to_do:
            task.status = TASK_STATUS_SCHEDULED

            logger.debug("Pushing task %s to the worker queue.", task.id)

            try:
                self.task_list.update(
                    id=task.id,
                    status=task.status
                )
            except StorageEngineError as e:
                logger.error(str(e))
            else:
                self.task_queue.put(task, False)

        to_redo = self.task_list.list_to_do(
            (TASK_STATUS_DONE | TASK_STATUS_FAILED | TASK_STATUS_ABORTED),
            now,
            redo=True
        )

        for task in to_redo:
            task.status = TASK_STATUS_SCHEDULED
            task.start_datetime = now
            task.stop_datetime = None
            task.output = ''

            logger.debug("Pushing task %s to the worker queue.", task.id)

            try:
                self.task_list.update(
                    id=task.id,
                    status=task.status,
                    start_datetime=task.start_datetime,
                    stop_datetime=task.stop_datetime,
                    output=task.output,
                )
            except StorageEngineError as e:
                logger.error(str(e))
            else:
                self.task_queue.put(task, False)

    def handle_message(self, message):
        if message.type == MSG_TYPE_TASK_NEW:
            # New task
            try:
                task_id = self.task_list.push(message.content)
                return Message(MSG_TYPE_RESP, {'id': task_id})
            except KeyError:
                return Message(MSG_TYPE_ERROR,
                               {'error': 'Task id already exists'})
            except StorageEngineError as e:
                logger.error(str(e))
                return Message(MSG_TYPE_ERROR, {'error': str(e)})

        elif message.type == MSG_TYPE_TASK_STATUS:
            # task status update
            status = message.content['status']
            # special case when task's status is TASK_STATUS_CANCELD, we dont'
            # want to change it's state.
            try:
                t = self.task_list.get(message.content['task_id'])
                if t.status & TASK_STATUS_CANCELED:
                    status = t.status

                self.task_list.update(
                    t.id,
                    status=status,
                    output=message.content.get('output', None),
                    stop_datetime=message.content.get('stop_datetime', None),
                )
            except StorageEngineError as e:
                logger.error(str(e))
                return Message(MSG_TYPE_ERROR, {'error': str(e)})
            else:
                return Message(MSG_TYPE_RESP, {'id': t.id})

        elif message.type == MSG_TYPE_TASK_LIST:
            # task list
            try:
                return list(self.task_list.list())
            except StorageEngineError as e:
                logger.error(str(e))
                return Message(MSG_TYPE_ERROR, {'error': str(e)})

        elif message.type == MSG_TYPE_TASK_ABORT:
            # task abortation
            t = Task(id=message.content['task_id'],
                     status=TASK_STATUS_ABORT)
            self.task_queue.put(t)

        elif message.type == MSG_TYPE_TASK_CANCEL:
            # task cancellation
            # first, we need to change its status and stop_datetime
            try:
                self.task_list.update(
                    message.content['task_id'],
                    status=TASK_STATUS_CANCELED,
                    stop_datetime=datetime.utcnow(),
                )
            except StorageEngineError as e:
                logger.error(str(e))
            else:
                # send the cancelation order to WP
                t = Task(id=message.content['task_id'],
                         status=TASK_STATUS_CANCELED)
                self.task_queue.put(t)
        elif message.type == MSG_TYPE_CONTEXT:
            # context update
            if type(message.content) == dict:
                for k, v in message.content.items():
                    self.set_context(k, v)
                return Message(MSG_TYPE_RESP, self.get_context())
            else:
                return Message(MSG_TYPE_ERROR, 'Unvalid type')


class SchedulerService(Service):
    # Adapter from taskmanager.Scheduler to Service

    def __init__(self, task_queue, event_queue, **kw):
        super(SchedulerService, self).__init__(**kw)
        self.task_queue = task_queue
        self.event_queue = event_queue
        self.scheduler = None
        self.task_list_engine = None

    def apply_config(self):
        # Setup scheduler as soon as configuration is loaded, before
        # plugins, so that tasklist is created before plugins.
        if not self.scheduler:
            self.scheduler = Scheduler(
                address=os.path.join(
                    self.app.config.temboard.home, '.tm.socket'),
                authkey=None)
            self.scheduler.task_queue = self.task_queue
            self.scheduler.event_queue = self.event_queue
            self.scheduler.task_list_engine = self.task_list_engine
            self.scheduler.setup_task_list()

    def setup(self):
        if os.path.exists(self.scheduler.address):
            os.unlink(self.scheduler.address)

        self.scheduler.setup()

    def serve1(self):
        self.scheduler.serve1()

    def add(self, workerset):
        if not self.is_my_process:
            return

        for task in workerset.list_tasks():
            try:
                task_from_db = self.scheduler.task_list.get(task.id)

                if task_from_db:
                    # If we've found the task in the DB, we'd keep its
                    # stop_datetime and status before overwriting it because
                    # we do not want this task to be re scheduled if it's not
                    # necessary.
                    task.stop_datetime = task_from_db.stop_datetime
                    task.status = task_from_db.status

                    self.scheduler.task_list.rm(task.id)
                    logger.debug("Overwriting task %s.", task.id)

                self.scheduler.task_list.push(task)
            except StorageEngineError as e:
                logger.error(str(e))

    def remove(self, workerset):
        if not self.is_my_process:
            return

        for task in workerset.list_tasks():
            try:
                self.scheduler.task_list.rm(task.id)
            except StorageEngineError as e:
                logger.error(str(e))

    def schedule_task(
            self, worker_name, id=None, options=None, start=None,
            redo_interval=None, expire=3600):
        message = Message(MSG_TYPE_TASK_NEW, Task(
            id=id,
            worker_name=worker_name,
            options=options,
            start_datetime=start or datetime.utcnow(),
            redo_interval=redo_interval,
            expire=expire,
        ))

        conn = Client(
            ensure_str(self.scheduler.address), self.scheduler.authkey)
        conn.send(message)
        res = conn.recv()
        conn.close()
        return res

    def can_schedule(self):
        return os.path.exists(self.scheduler.address)


class WorkerPool(object):
    trace = False

    def __init__(self, task_queue, event_queue, setproctitle=None):
        self.thread = None
        self.task_queue = task_queue
        self.event_queue = event_queue
        self.workers = {}
        self.setproctitle = setproctitle
        self.perf = None

    def _abort_job(self, task_id):
        for workername in self.workers:
            for job in self.workers[workername]['pool']:
                if job['id'] == task_id:
                    logger.debug("Process pid=%s is going to be killed"
                                 % job['process'])
                    job['process'].terminate()
                    return True
        return False

    def _rm_task_worker_queue(self, task_id):
        for workername in self.workers:
            for t in self.workers[workername]['queue']:
                if t.id == task_id:
                    self.workers[workername]['queue'].remove(t)
                    logger.debug("Task %s removed from queue" % t.id)
                    return True
        return False

    def setup(self):
        if self.perf:
            self.perf['fork'] = 0

    def add(self, worker):
        if worker['name'] in self.workers:
            raise Exception("Worker %s already registered." % worker['name'])

        self.workers[worker['name']] = {
            'queue': deque(),
            'pool_size': worker['pool_size'],
            'module': worker['module'],
            'function': worker['function'],
            'pool': []
        }

    def serve1(self):
        # check running jobs state
        self.check_jobs()
        # start new jobs
        self.start_jobs()

        try:
            t = self.task_queue.get(timeout=0.1)
        except Empty:
            return

        if t.status & TASK_STATUS_SCHEDULED:
            logger.debug("Got task %s for worker '%s' queue"
                         % (t.id, t.worker_name))
            self.workers[t.worker_name]['queue'].appendleft(t)
            # Update task status
            self.event_queue.put(
                Message(
                    MSG_TYPE_TASK_STATUS,
                    {
                        'task_id': t.id,
                        'status': TASK_STATUS_QUEUED,
                    }
                )
            )
        if t.status & TASK_STATUS_ABORT:
            self._abort_job(t.id)
        if t.status & TASK_STATUS_CANCELED:
            # Task cancellation includes 2 things:
            # - remove the task from workers queue if present (job not
            # yet running)
            # - abort running job executing the task if any

            # Looking up into workers jobs first, there is more chance
            # to find the task here.
            if not self._abort_job(t.id):
                # If not aborted, task has been queued
                self._rm_task_worker_queue(t.id)

    def exec_worker(self, module, function, out, *args, **kws):
        # Function wrapper around worker function
        try:
            # Reset signal handlers.
            signal.signal(signal.SIGABRT, signal.SIG_DFL)
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            signal.signal(signal.SIGTERM, signal.SIG_DFL)

            fun = getattr(sys.modules[module], function)

            modfun = "%s.%s" % (module, fun.__name__)
            logger.debug("Starting new job for %s", modfun)
            if self.setproctitle:
                self.setproctitle('task %s' % (modfun))
            perf = PerfCounters.setup(service='task', task=modfun)
            if perf:
                perf.schedule()

            res = fun(*args, **kws)
            # Put function result into output queue as a Message
            out.put(Message(MSG_TYPE_RESP, res))
        except UserError as e:
            logger.critical("%s", e)
        except Exception as e:
            e = Exception("%s: %s" % (type(e), e))
            out.put(Message(MSG_TYPE_ERROR, e))
            logger.exception(e)
        except KeyboardInterrupt:
            logger.error("KeyboardInterrupt")
        if perf:
            perf.run()

    def start_jobs(self):
        # Execute Tasks
        for name, worker in self.workers.items():
            while len(self.workers[name]['pool']) < worker['pool_size']:
                try:
                    t = worker['queue'].pop()
                    # Queue used to get worker function return
                    out = Queue()
                    p = Process(
                            target=self.exec_worker,
                            args=(worker['module'], worker['function'], out),
                            kwargs=t.options,
                        )
                    p.start()

                    if self.perf:
                        self.perf['fork'] += 1

                    self.workers[name]['pool'].append(
                            {'id': t.id, 'process': p, 'out': out}
                    )
                    # Update task status
                    self.event_queue.put(
                        Message(
                            MSG_TYPE_TASK_STATUS,
                            {
                                'task_id': t.id,
                                'status': TASK_STATUS_DOING,
                            }
                        )
                    )
                except IndexError:
                    break

    def check_jobs(self):
        # Check jobs process state for each worker
        for name, worker in self.workers.items():
            for job in worker['pool']:
                if not job['process'].is_alive():
                    # Dead process case
                    logger.debug("Job %s terminated.", job['id'])
                    try:
                        # Fetch the message from job's output queue
                        message_out = job['out'].get(False)
                    except Empty:
                        message_out = None
                    if self.trace:
                        logger.debug("Job output : %s" % message_out)
                    # Close job's output queue
                    job['out'].close()
                    # join the process
                    job['process'].join()

                    # Let's build the message we'll have to send to scheduler
                    # for the update of task's status.
                    task_stop_dt = datetime.utcnow()
                    if job['process'].exitcode == 0:
                        if message_out and \
                                message_out.type[0] == MSG_TYPE_RESP:
                            task_status = TASK_STATUS_DONE
                        else:
                            # when an exception is raised from the worker
                            # function
                            task_status = TASK_STATUS_FAILED
                    elif job['process'].exitcode < 0:
                        # process killed
                        task_status = TASK_STATUS_ABORTED
                    else:
                        task_status = TASK_STATUS_FAILED
                    task_output = None
                    if message_out:
                        task_output = message_out.content

                    # Update task status
                    self.event_queue.put(
                        Message(
                            MSG_TYPE_TASK_STATUS,
                            {
                                'task_id': job['id'],
                                'status': task_status,
                                'output': task_output,
                                'stop_datetime': task_stop_dt
                            }
                        )
                    )

                    # Finally, remove the job from the pool
                    self.workers[name]['pool'].remove(job)

    def abort_jobs(self):
        # abort all running jobs
        for name, worker in self.workers.items():
            for job in worker['pool']:
                process = job.get('process')
                if process.is_alive():
                    process.terminate()
                    logger.debug("Job %s has been terminated" % job['id'])
                    # Close job's output queue
                    job['out'].close()
                    # join the process
                    process.join()
                    self.workers[name]['pool'].remove(job)


class WorkerSet(list):
    def register(self, pool_size=1):
        def register(f):
            def defer(app, **kw):
                logger.debug("Scheduling %s.", f.__name__)
                return app.scheduler.schedule_task(
                    f.__name__, options=kw, expire=0)
            f.defer = defer

            f._tm_worker = make_worker_definition(f, pool_size)
            if f not in self:
                self.append(f)
            return f
        return register

    def schedule(self, id=None, redo_interval=None, **options):
        def register(f):
            if f not in self:
                self.append(f)
            f._tm_task = Task(
                id=id,
                options=options,
                redo_interval=redo_interval,
                worker_name=f.__name__,
            )
            return f
        return register

    def list_tasks(self):
        for function in self:
            task = getattr(function, '_tm_task', None)
            if task:
                yield task


class WorkerPoolService(Service):
    # Adapter from taskmanager.WorkerPool to Service

    def __init__(self, task_queue, event_queue, **kw):
        super(WorkerPoolService, self).__init__(**kw)
        self.worker_pool = WorkerPool(
            task_queue, event_queue, self.setproctitle)

    def setup(self):
        self.worker_pool.perf = self.perf
        self.worker_pool.setup()

    def serve1(self):
        try:
            self.worker_pool.serve1()
        except Exception:
            logger.exception("Unhandled error in worker:")
            logger.error("Not stopping worker process.")

    def create_task_function_app_wrapper(self, function):
        @functools.wraps(function)
        def wrapper(*a, **kw):
            return function(self.app, *a, **kw)
        wrapper._tm_function = function
        return wrapper

    def add(self, workerset):
        if not self.is_my_process:
            return

        for function in workerset:
            conf = function._tm_worker
            wrapper = self.create_task_function_app_wrapper(function)

            # Inject wrapper in module so taskmanager will find it.
            mod = sys.modules[conf['module']]
            wrapper_name = '_tm_wrapper_' + function.__name__
            setattr(mod, wrapper_name, wrapper)
            conf['function'] = wrapper_name

            # Add to current workers
            logger.debug("Register worker %s", conf['name'])
            self.worker_pool.add(conf)

    def remove(self, workerset):
        if not self.is_my_process:
            return

        for function in workerset:
            conf = function._tm_worker
            logger.debug("Disable worker %s", conf['name'])
            self.worker_pool.workers.pop(conf['name'], None)

    def sigterm_handler(self, *a):
        logger.info("Aborting jobs on SIGTERM.")
        self.worker_pool.abort_jobs()
        super(WorkerPoolService, self).sigterm_handler(*a)


class FlushTasksMixin(object):
    # Shared code for flush-task command

    def define_arguments(self, parser):
        parser.add_argument(
            "--force",
            action='store_true', default=False,
            help="Force overwriting existing files.",
        )

    def main(self, args):
        if self.app.scheduler.can_schedule():
            if args.force:
                logger.warning(
                    "Scheduler socket exists. Is temBoard scheduler running?")
            else:
                logger.error("Scheduler socket exists. Use --force to bypass.")
                raise UserError("You must stop temBoard before flushing tasks")

        count = self.app.scheduler.task_list_engine.flush()
        logger.info("Flushed %s tasks.", count)
        return 0


class RunTaskMixin(object):
    # Shared code to execute a single task, for runtask commands.
    #
    # To combines with toolkit.app.SubCommand

    def define_arguments(self, parser):
        parser.description = dedent("""\

        Run a task foreground. Some tasks won't work foreground because they
        requires task manager processes.

        Use this only for testing, debugging and development.

        Note that our home-made background task implementation uses a different
        semantic that state of the art background task implementation:

        - a task function is called a worker
        - a worker process is called workerpool
        - a message is called a task
        - the message broker is called taskmanager

        """)

        parser.add_argument(
            'worker_name',
            metavar='WORKER',
            help=(
                "Global name of the worker function name to execute."
                " Use ? to list available workers."),
        )

        parser.add_argument(
            'worker_args', nargs='*',
            metavar='ARG',
            default=[],
            help="Worker arguments as Python literals.",
        )

    def iter_workers(self):
        for name, config in self.app.worker_pool.worker_pool.workers.items():
            mod = sys.modules[config['module']]
            fn = getattr(mod, config['function'])
            yield fn

    def compute_worker_args(self, workers, args):
        needles = (args.worker_name, args.worker_name + '_worker')
        for worker in workers:
            if worker.__name__ in needles:
                break
        else:
            raise UserError("Unknown worker %s." % args.worker_name)

        worker_args = []
        for arg in args.worker_args:
            try:
                arg = literal_eval(arg)
            except Exception:
                logger.debug("Unknown literal %s, using as raw string.", arg)
            worker_args.append(arg)

        return worker, worker_args

    def print_workers(self, workers):
        for name in sorted(fn.__name__ for fn in workers):
            print(name)



# Simple live performance counters in logfmt.

import logging
import os
import signal


logger = logging.getLogger(__name__)
SC_CLK_TCK = os.sysconf('SC_CLK_TCK')


class PerfCounters(dict):
    @classmethod
    def setup(cls, **defaults):
        if 'PERF' not in os.environ:
            return
        return cls(**defaults)

    def __init__(self, **defaults):
        defaults = dict(
            io_rchar=0,
            io_wchar=0,
            stime=0,
            utime=0,
            vsize=0,
            pid=os.getpid(),
            load1=0,
            load5=0,
            load15=0,
            **defaults
        )
        super(PerfCounters, self).__init__(**defaults)

        try:
            self.delay = int(os.environ['PERF'])
        except (KeyError, ValueError):
            self.delay = 60

        self.delay = max(15, self.delay)
        logger.debug("Scheduling perf counters each %s seconds.", self.delay)

    def __str__(self):
        return ' '.join([
            '%s=%s' % i
            for i in sorted(self.items())
        ])

    def run(self):
        try:
            self.snapshot()
        except Exception as e:
            logger.error("Failed to read performances metrics: %s", e)
            return
        logger.debug("%s", self)
        self.schedule()

    def sigalrm_handler(self, *a):
        self.run()

    def schedule(self):
        signal.alarm(self.delay)

    def snapshot(self):
        # I/O
        needles = ['rchar', 'wchar']
        with open('/proc/self/io') as fo:
            for k, v in parse(fo):
                if k not in needles:
                    continue
                self['io_%s' % k] = v
                if not needles:
                    break

        # PROCESS
        with open('/proc/self/stat') as fo:
            stat = fo.read()

        # Fields are documented in proc(5).
        fields = stat.split()

        utime = int(fields[13]) / SC_CLK_TCK
        self['utime'] = utime

        stime = int(fields[14]) / SC_CLK_TCK
        self['stime'] = stime

        vsize = int(fields[22])
        self['vsize'] = vsize

        # GLOBAL LOAD AVG
        with open('/proc/loadavg') as fo:
            loadavg = fo.read()
        self['load1'], self['load5'], self['load15'], _ = loadavg.split(' ', 3)


def parse(lines):
    for line in lines:
        try:
            k, v = line.split(': ')
        except ValueError:
            continue
        yield k, v.strip()



import logging
import os
import datetime
import getpass
from argparse import _VersionAction
from socket import getfqdn
from textwrap import dedent

from ..core import workers
from ..discover import Discover, inspect_versions
from ..queries import QUERIES
from ..status import Status
from ..toolkit.configuration import OptionSpec
from ..toolkit.errors import UserError
from ..web import HTTPDService
from ..postgres import Postgres
from ..toolkit import taskmanager, validators as v
from ..toolkit.app import BaseApplication, define_core_arguments
from ..toolkit.configuration import MergedConfiguration
from ..toolkit.proctitle import ProcTitleManager
from ..toolkit.signing import load_public_key
from ..toolkit.tasklist.sqlite3_engine import TaskListSQLite3Engine
from ..notification import NotificationMgmt
from ..version import __version__


logger = logging.getLogger('temboardagent.scripts.agent')


class TemboardAgentApplication(BaseApplication):
    PROGRAM = "temboard-agent"
    VERSION = __version__

    DEFAULT_CONFIGFILES = [
        '/etc/temboard-agent/temboard-agent.conf',
        'temboard-agent.conf',
    ]
    DEFAULT_PLUGINS = [
        "activity",
        "administration",
        "dashboard",
        "maintenance",
        "monitoring",
        "pgconf",
        "statements",
    ]

    def __init__(self, *a, **kw):
        super(TemboardAgentApplication, self).__init__(*a, **kw)
        self.config = TemboardAgentConfiguration()

    def main(self, argv, environ):
        parser = self.create_parser(
            description=dedent("""\
            temBoard agent %s.

            COMMAND is optional. Default command is `serve`, the combined
            service. See available commands below.

            """) % __version__,
        )
        self.define_arguments(parser)
        args = parser.parse_args(argv)

        command_name = getattr(args, 'command_fullname', 'serve')
        command = self.commands[command_name]

        setproctitle = ProcTitleManager(prefix='temboard-agent: ')
        setproctitle.setup()

        task_queue = taskmanager.Queue()
        event_queue = taskmanager.Queue()

        self.worker_pool = taskmanager.WorkerPoolService(
            app=self, setproctitle=setproctitle, name='worker pool',
            task_queue=task_queue, event_queue=event_queue)
        self.services.append(self.worker_pool)
        self.worker_pool.add(workers)

        self.scheduler = taskmanager.SchedulerService(
            app=self, setproctitle=setproctitle, name='scheduler',
            task_queue=task_queue, event_queue=event_queue)
        self.services.append(self.scheduler)

        self.httpd = HTTPDService(
            self, setproctitle=setproctitle, name='web',
        )

        self.bootstrap(args=args, environ=environ, service=command.is_service)
        self.log_versions()
        config = self.config

        self.discover = Discover(self)
        self.discover.read()
        self.status = Status(self)

        # TaskList engine setup must be done before we load the plugins
        self.scheduler.task_list_engine = TaskListSQLite3Engine(
            os.path.join(config.temboard['home'], 'agent_tasks.db')
        )

        self.apply_config()

        if '.' not in self.config.temboard.hostname:
            logger.warning(
                "Hostname %s is not a FQDN.",
                self.config.temboard.hostname
            )

        cluster_name = self.discover.data['postgres'].get('cluster_name')
        if cluster_name:
            setproctitle.prefix += cluster_name + ': '

        QUERIES.load()

        self.start_datetime = datetime.datetime.now()
        self.reload_datetime = None
        self.pid = os.getpid()
        self.user = getpass.getuser()

        self.bootstrap_plugins()

        # Boostraping action logs table
        NotificationMgmt.bootstrap(config)

        return command.main(args)

    def apply_config(self):
        self.postgres = Postgres(app=self, **self.config.postgresql)
        self.postgres.connection_lost_observers.append(
            self.discover.connection_lost)
        return super().apply_config()

    def bootstrap_plugins(self):
        for plugin_name, plugin in self.plugins.items():
            if hasattr(plugin, 'bootstrap'):
                logger.debug("Boostraping plugin %s", plugin_name)
                plugin.bootstrap()

    def check_compatibility(self, pg_version):
        # check for compatibility with plugins
        for name, plugin in self.plugins.items():
            if pg_version < plugin.PG_MIN_VERSION[0]:
                logger.error(
                    "%s plugin is incompatible with Postgres below %s",
                    name, plugin.PG_MIN_VERSION[1],
                )

    def define_arguments(self, parser):
        define_core_arguments(parser)
        parser.add_argument(
            '-V', '--version',
            action=VersionAction,
            help='show version and exit',
        )
        parser.add_argument(
            '-d', '--daemon',
            action='store_true', dest='temboard_daemonize',
            help="Run in background.",
        )
        parser.add_argument(
            '-p', '--pid-file',
            action='store', dest='temboard_pidfile',
            help="PID file.",
        )
        super(TemboardAgentApplication, self).define_arguments(parser)

    def init_specs(self, app_specs):
        specs = super().init_specs(app_specs)

        def add_specs(*new_specs):
            for spec in new_specs:
                specs[str(spec)] = spec

        # These are *core* because they are needed to load plugins.
        s = 'postgresql'
        add_specs(
            OptionSpec(
                s, 'host', default='/var/run/postgresql', validator=v.dir_),
            OptionSpec(s, 'port', default=5432, validator=v.port),
            OptionSpec(s, 'user', default='postgres'),
            OptionSpec(s, 'password'),
            OptionSpec(s, 'dbname', default='postgres'),
        )

        return specs

    def core_specs(self):
        for spec in super().core_specs():
            yield spec

        for name, spec in self.config_specs.items():
            if name.startswith('postgresql_'):
                yield spec

    def log_versions(self):
        versions = inspect_versions()
        logger.debug(
            "Running on %s %s.",
            versions['distname'], versions['distversion'])
        logger.debug(
            "Using Python %s (%s).",
            versions['python'], versions['pythonbin'])
        logger.debug(
            "Using libpq %s, Psycopg2 %s.",
            versions['libpq'], versions['psycopg2'],
        )

    def reload(self):
        super().reload()
        self.reload_datetime = datetime.datetime.now()


class TemboardAgentConfiguration(MergedConfiguration):
    def __init__(self, *a, **kw):
        super(TemboardAgentConfiguration, self).__init__(*a, **kw)
        self._signing_key = None

    @property
    def signing_key(self):
        # Lazy load signing key.
        if not self._signing_key:
            self.load_signing_key()
        return self._signing_key

    def load_signing_key(self):
        path = self.temboard.signing_public_key
        logger.debug("Loading signing key from %s.", path)
        try:
            fo = open(path, 'rb')
        except OSError as e:
            raise UserError("Failed to load signing key: %s" % e)

        with fo:
            self._signing_key = load_public_key(fo.read())


class VersionAction(_VersionAction):
    fmt = dedent("""\
    temBoard agent %(temboard)s (%(temboardbin)s)
    System %(distname)s %(distversion)s
    Python %(python)s (%(pythonbin)s)
    bottle %(bottle)s
    cryptography %(cryptography)s
    libpq %(libpq)s
    psycopg2 %(psycopg2)s
    """)

    def __call__(self, parser, *_):
        print((self.fmt % inspect_versions()).strip())
        parser.exit()


def list_options_specs():
    # Generate each option specs.
    section = 'temboard'
    yield OptionSpec(section, 'ui_url', validator=v.url)
    yield OptionSpec(section, 'daemonize', default=False)
    yield OptionSpec(section, 'pidfile', default='/run/temboard-agent.pid')
    yield OptionSpec(
        section, 'address', default='0.0.0.0', validator=v.address)
    yield OptionSpec(section, 'port', validator=v.port, default=2345)
    yield OptionSpec(
        section, 'ssl_cert_file',
        default=OptionSpec.REQUIRED, validator=v.file_)
    yield OptionSpec(
        section, 'ssl_key_file',
        default=OptionSpec.REQUIRED, validator=v.file_)
    yield OptionSpec(section, 'ssl_ca_cert_file', validator=v.file_)
    yield OptionSpec(
        section, 'signing_public_key', default='signing-public.pem',
        validator=v.path)
    yield OptionSpec(section, 'key')
    yield OptionSpec(section, 'hostname', default=getfqdn(), validator=v.fqdn)
    home = os.environ.get('HOME', '/var/lib/temboard-agent')
    yield OptionSpec(section, 'home', default=home, validator=v.writeabledir)


app = TemboardAgentApplication(specs=list_options_specs())



# coding: utf-8
#
# This module implements background service management.
#
# Background service management features:
# - Proper starting and killing of background service.
# - proctitle definition.
# - SIGTERM and SIGHUP propagation.
# - SIGCHLD management. No zombie.
# - Child suicide on parent death.
#
# Two classes help manage multiple services: Service and ServicesManager.
#
# Each long running process must have it's own Service instance. Service class
# is used both to manage current process service as well as reference to
# manipulate long-runnning child process.
#
# ServicesManager class is kind of an init system for the main process. It is
# responsible of starting, killing and propagating signals to child processes.
# Service instance of main process service must have a reference to
# ServicesManager.

import logging
import os
import signal
import sys
from multiprocessing import Process
from time import sleep

from .errors import UserError
from .perf import PerfCounters


logger = logging.getLogger(__name__)


class Service(object):
    # Manage long running process. This include setup, signal management and
    # loop.
    #
    # There is two kind of services : main service and child services. Main
    # service is responsible to propagate signals to children with
    # ServicesManager.

    def __init__(self, app, name=None, services=None, setproctitle=None):
        self.app = app
        self.name = name
        self.logname = name or u'service'
        # Must be None for children or ServicesManager instance for main
        # service. Used to propagate signals. See reload() method.
        self.services = services
        self.setproctitle = setproctitle

        self.parentpid = None
        # Once the process is forked to run the service loop, we still use this
        # object in parent process to manage the child process. So we flag here
        # whether the service has forked in it's own process. Must be updated
        # in parent process once the service is forked. See
        # ServicesManager.start().
        self.is_my_process = True

    def __str__(self):
        return self.logname

    def __unicode__(self):
        return self.logname

    def __enter__(self):
        self.sigchld = False
        if self.services:
            signal.signal(signal.SIGCHLD, self.sigchld_handler)
        signal.signal(signal.SIGHUP, self.sighup_handler)
        signal.signal(signal.SIGTERM, self.sigterm_handler)
        self.sighup = False
        if self.perf:
            signal.signal(signal.SIGALRM, self.perf.sigalrm_handler)

    def __exit__(self, *a):
        signal.signal(signal.SIGCHLD, signal.SIG_DFL)
        signal.signal(signal.SIGHUP, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)

    def check_parent_running(self):
        if self.parentpid is None:
            # If no parentpid, we are the main service. We are running.
            return True

        try:
            os.kill(self.parentpid, 0)
            return True
        except OSError:
            return False

    def sigchld_handler(self, *a):
        self.sigchld = True

    def sighup_handler(self, *a):
        self.sighup = True

    def sigterm_handler(self, *a):
        logger.info(u"%s terminated.", self)
        sys.exit(0)

    def apply_config(self):
        pass

    def run(self):
        if self.name and self.setproctitle:
            self.setproctitle(self.name)
        logger.info(u"Starting %s.", self)
        self.perf = PerfCounters.setup(service=self.logname.replace(' ', '-'))
        if self.perf:
            self.perf.run()

        self.setup()

        try:
            self.serve()
        except KeyboardInterrupt:
            logger.info(u"%s interrupted.", self)
            sys.exit(0)

    def serve(self):
        with self:
            logger.debug(u"Entering %s loop.", self)
            while True:
                if not self.check_parent_running():
                    logger.warn(
                        u"Parent process %d is dead. Committing suicide.",
                        self.parentpid)
                    sys.exit(1)

                if self.sigchld:
                    self.sigchld = False
                    if self.services:
                        self.services.check()

                if self.sighup:
                    self.sighup = False
                    self.reload()

                self.serve1()

    def reload(self):
        self.app.reload()
        if self.services:
            self.services.reload()

    def setup(self):
        # This method is called once before looping to prepare the service:
        # bind address, setup SSL, etc.
        pass

    def serve1(self):
        # This method is called by the loop and must serve one request/task.
        # This method should not block for too long waiting for work to be
        # done. Reload is applied between two calls of this method.
        raise NotImplementedError


class ServicesManager(object):
    # Manage child services : starting in background, tracking PID, replicating
    # signals, checking status, stopping and killing.
    #
    # Add a service with services_manager.add(Service(…)).
    #
    # As a context manager, services are started on enter and stopped-killed on
    # exit.

    def __init__(self):
        self.processes = []
        self.pid = os.getpid()

    def __enter__(self):
        self.start()

    def __exit__(self, *a):
        self.stop()
        logger.debug(u"Waiting background services.")
        sleep(0.125)
        self.kill()

    def add(self, service):
        service.parentpid = self.pid
        self.processes.append(
            (service, Process(target=service.run, name=service.name)))

    def start(self):
        for service, process in self.processes:
            process.start()
            service.is_my_process = False

    def reload(self):
        for _, process in self.processes:
            os.kill(process.pid, signal.SIGHUP)

    def check(self):
        for i in self.processes[:]:
            _, p = i
            logger.debug(u"Checking child %s (%s).", p.name, p.pid)
            if not p.is_alive():
                logger.debug("%s (%s) is dead.", p.name, p.pid)
                self.processes.remove(i)
                msg = u"Child %s (%s) died." % (p.name, p.pid)
                raise UserError(msg)

    def stop(self):
        for _, process in self.processes:
            process.terminate()

    def kill(self, timeout=5, step=0.5):
        while timeout > 0:
            processes = [p for _, p in self.processes if p.is_alive()]
            if not processes:
                break
            sleep(step)
            timeout -= step

        for process in processes:
            logger.warning(u"Killing %s pid=%s.", process, process.pid)
            os.kill(process.pid, signal.SIGKILL)



class Status(object):
    def __init__(self, app):
        self.app = app
        self.data = dict(
            postgres=dict(available=True),
            temboard=dict(),
            system=dict(),
        )



# coding: utf-8
from __future__ import absolute_import

import http.client
import json
import logging
import ssl
from datetime import datetime
try:
    from datetime import timezone
except ImportError:  # PY2
    timezone = None
from time import time

from .utils import ensure_bytes
from .errors import TemboardError
from .pycompat import PY2, HTTPError

try:
    ConnectionError = ConnectionError
except NameError:  # python2
    from socket import error as ConnectionError


logger = logging.getLogger(__name__)


class TemboardHTTPError(TemboardError):
    def __init__(self, response):
        self.response = response
        self._message = None
        super(TemboardHTTPError, self).__init__(response.status, self.message)

    @property
    def message(self):
        if not self._message:
            try:
                json = self.response.json()
                self._message = json['error']
            except Exception:
                self._message = self.response.reason
        return self._message

    def __str__(self):
        return '%s: %s' % (self.response.status, self.message)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self)


class TemboardClient(object):
    ConnectionError = ConnectionError
    Error = TemboardHTTPError

    log_headers = False

    @classmethod
    def factory(cls, config, host, port, scheme='https'):
        return cls(
            host, port, scheme=scheme,
            ca_cert_file=config.temboard.ssl_ca_cert_file,
        )

    def __init__(self, host, port, ca_cert_file=None, scheme='https'):
        """ If ca_cert_file is None, HTTPS connection is unverified. """
        self.scheme = scheme
        self.host = host
        self.port = port
        self.ca_cert_file = ca_cert_file
        self._ssl_context = None
        self._cookies = set()

    def __repr__(self):
        return '<%s %s://%s:%s %s>' % (
            self.__class__.__name__,
            self.scheme, self.host, self.port,
            'verified' if self.ca_cert_file else 'unverified',
        )

    @property
    def ssl_context(self):
        if self._ssl_context is None:
            self._ssl_context = ssl.create_default_context(
                cafile=self.ca_cert_file,
            )
            if self.ca_cert_file:
                self._ssl_context.verify_mode = ssl.CERT_REQUIRED
            else:

                self._ssl_context.check_hostname = False
                self._ssl_context.verify_mode = ssl.CERT_NONE
        return self._ssl_context

    def request(self, method, path, headers=None, body=None):
        hostport = '%s:%s' % (self.host, self.port)
        fullurl = '%s://%s%s' % (self.scheme, hostport, path)

        logger.debug("Requesting %s %s.", method, fullurl)

        headers = headers or {}
        if self._cookies:
            headers.setdefault('Cookie', "\n".join(self._cookies))

        if body:
            headers['Content-Type'] = 'application/json'

        if hasattr(body, 'pop'):  # list or dict
            body = json.dumps(body)

        if body is not None:
            body = ensure_bytes(body)

        if 'https' == self.scheme:
            conn = http.client.HTTPSConnection(
                self.host, self.port, context=self.ssl_context, timeout=30,
            )
        else:
            conn = http.client.HTTPConnection(
                self.host, self.port, timeout=30,
            )
        conn.response_class = TemboardResponse

        if self.log_headers:
            for name, value in sorted(headers.items()):
                logger.debug(">>> %s: %s", name, value)

        start_time = time()
        conn.request(method, path, body, headers)
        response = conn.getresponse()
        duration = time() - start_time
        response.path = path

        if self.log_headers:
            for name, value in sorted(response.headers.items()):
                logger.debug("<<< %s: %s", name, value)

        if response.headers['set-cookie']:
            cookie = response.headers['set-cookie']
            logger.debug("Registering cookie %.16s... in session.", cookie)
            self._cookies.add(cookie)

        logger.debug(
            "Response from %s in %.3fs: %s.", hostport, duration, response)

        return response

    def get(self, path, headers=None):
        return self.request('GET', path, headers)

    def post(self, path, body, headers=None):
        return self.request('POST', path, headers, body)


class TemboardResponse(http.client.HTTPResponse):
    # Extensions to HTTPResponse, inspired by httpx

    if PY2:
        @property
        def headers(self):
            return dict(self.getheaders())

    def __str__(self):
        return '%s %s' % (self.status, self.reason)

    def __repr__(self):
        return '<%s %s %s -> %s %s %s>' % (
            self.__class__.__name__,
            self._method, self.path,
            self.status, self.reason,
            'closed' if self.isclosed() else 'opened',
        )

    def raise_for_status(self):
        if self.status >= 400:
            raise TemboardHTTPError(self)
        elif self.status >= 300:
            raise HTTPError(self.status, self.reason)

    def json(self):
        return json.loads(self.read().decode('utf-8'))


def format_date(date=None):
    if not date:
        date = datetime.utcnow()
        date = date.replace(tzinfo=timezone.utc)
    if date.tzinfo is None:
        raise Exception("Refusing to format naïve datetime.")
    if date.tzinfo is not timezone.utc:
        date = date.astimezone(timezone.utc)
    return date.strftime("%Y%m%dT%H%M%SZ")



import logging

from ..toolkit.app import SubCommand
from .app import app


logger = logging.getLogger(__name__)


@app.command
class RunScript(SubCommand):
    """ Run a Python script. """

    def define_arguments(self, parser):
        parser.add_argument(
            'script', metavar='PATH',
            help="Path to script file.",
        )

        super(RunScript, self).define_arguments(parser)

    def main(self, args):
        with open(args.script) as fo:
            source = fo.read()
        code = compile(source, args.script, 'exec')

        g = globals()
        g['app'] = self.app

        exec(code, g, g)



#
# Configure and extend psycopg2 for temBoard
#
# - Use float for decimal.
# - dict as default row type
# - per query row_factory
# - connection helpers to hide cursor.
#

import ctypes
import logging
import re
from contextlib import closing
from time import sleep

from psycopg2 import connect
from psycopg2 import Error as Psycopg2Error
from psycopg2.pool import ThreadedConnectionPool
import psycopg2.extensions

from .toolkit.versions import load_libpq


logger = logging.getLogger(__name__)


# See https://www.psycopg.org/docs/faq.html#faq-float
DEC2FLOAT = psycopg2.extensions.new_type(
    psycopg2.extensions.DECIMAL.values,
    'DEC2FLOAT',
    lambda value, curs: float(value) if value is not None else None)
psycopg2.extensions.register_type(DEC2FLOAT)


def pg_escape(in_string, escapee_char=r"'"):
    out_string = ''
    out_string += escapee_char
    out_string += re.sub(escapee_char, escapee_char * 2, in_string)
    out_string += escapee_char
    return out_string


class Postgres:
    # main object holding Postgres parameters and methods.

    def __init__(
            self, host=None, port=5432, user=None, password=None, dbname=None,
            app=None,
            **kw):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        # Compat with conninfo dict.
        if 'database' in kw:
            dbname = kw['database']
        self.dbname = dbname
        self.app = app
        self._server_version = None
        self.connection_lost_observers = []

    def __repr__(self):
        return '<{} on {}@{}:{}/{}>'.format(
            self.__class__.__name__,
            self.user, self.host, self.port, self.dbname,
        )

    def dbpool(self):
        return DBConnectionPool(self)

    def pool(self):
        return ConnectionPool(
            app=self.app,
            observers=self.connection_lost_observers,
            minconn=1, maxconn=2,
            **self.pqvars(),
        )

    def connect(self, database=None):
        kw = self.pqvars()
        if database:
            kw['dbname'] = database
        return closing(retry_connect(connect, self.app, **kw))

    def copy(self, **kw):
        defaults = dict(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            dbname=self.dbname,
            app=self.app,
        )
        kw = dict(defaults, **kw)
        return self.__class__(**kw)

    def pqvars(self, dbname=None):
        return dict(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            database=dbname or self.dbname,
            connection_factory=FactoryConnection,
            application_name='temboard-agent',
        )

    def notify_observers(self):
        for observer in self.connection_lost_observers:
            observer()


class DBConnectionPool:
    # Pool one connection per database.
    #
    # Not thread-safe.

    def __init__(self, postgres):
        self.postgres = postgres
        self.pool = dict()

    def getconn(self, dbname=None):
        dbname = dbname or self.postgres.dbname
        conn = self.pool.get(dbname)
        if conn and conn.pqstatus() == conn.CONNECTION_BAD:
            logger.debug("Recycling bad connection to db %s.", dbname)
            conn.close()
            del self.pool[dbname]
            conn = None

        if not conn:
            logger.debug("Opening connection to db %s.", dbname)
            pqvars = self.postgres.pqvars(dbname=dbname)
            conn = self.pool.setdefault(dbname, connect(**pqvars))

        return conn

    def closeall(self):
        for dbname in list(self.pool):
            conn = self.pool.pop(dbname)
            logger.debug("Closing pooled connection to %s.", dbname)
            conn.close()

    def __del__(self):
        self.closeall()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.closeall()


class FactoryConnection(psycopg2.extensions.connection):  # pragma: nocover
    CONNECTION_OK = 0
    CONNECTION_BAD = 1

    def pqstatus(self):
        try:
            self.libpq
        except AttributeError:
            self.libpq = load_libpq()
            self.libpq.PQstatus.argtypes = [ctypes.c_void_p]
            self.libpq.PQstatus.restype = ctypes.c_int

        return self.libpq.PQstatus(self.pgconn_ptr)

    def cursor(self, *a, **kw):
        row_factory = kw.pop('row_factory', None)
        kw['cursor_factory'] = FactoryCursor.make_factory(
            row_factory=row_factory)
        return super(FactoryConnection, self).cursor(*a, **kw)

    def execute(self, sql, *args):
        with self.cursor() as cur:
            cur.execute(sql, *args)

    def query(self, sql, *args, row_factory=None):
        with self.cursor(row_factory=row_factory) as cur:
            cur.execute(sql, *args)
            for row in cur.fetchall():
                yield row

    def queryone(self, sql, *args, row_factory=None):
        with self.cursor(row_factory=row_factory) as cur:
            cur.execute(sql, *args)
            return cur.fetchone()

    def queryscalar(self, sql, *args):
        return self.queryone(sql, *args, row_factory=scalar)


class FactoryCursor(psycopg2.extensions.cursor):  # pragma: nocover
    # Implement row_factory for psycopg2.

    @classmethod
    def make_factory(cls, row_factory=None):
        # Build a cursor_factory for psycopg2 connection.
        def factory(*a, **kw):
            kw['row_factory'] = row_factory
            return cls(*a, **kw)
        return factory

    def __init__(self, conn, name=None, row_factory=None):
        super(FactoryCursor, self).__init__(conn)
        if not row_factory:
            def row_factory(**kw):
                return kw
        self._row_factory = row_factory

    def fetchone(self):
        row = super(FactoryCursor, self).fetchone()
        kw = dict(zip([c.name for c in self.description], row))
        return self._row_factory(**kw)

    def fetchmany(self, size=None):
        for row in super(FactoryCursor, self).fetchmany(size):
            kw = dict(zip([c.name for c in self.description], row))
            yield self._row_factory(**kw)

    def fetchall(self):
        for row in super(FactoryCursor, self).fetchall():
            kw = dict(zip([c.name for c in self.description], row))
            yield self._row_factory(**kw)


def scalar(**kw):
    # Row factory for scalar.
    return next(iter(kw.values()))


class RetryManager(object):
    def __init__(self, pool):
        self.pool = pool
        self.conn = None
        self.retry = False

    def __call__(self):
        return self

    def __enter__(self):
        self.conn = self.pool.getconn()
        self.retry = False
        return self.conn

    def __exit__(self, exc_type, e, exc_tb):
        if isinstance(e, Psycopg2Error):
            if e.pgcode is None:
                logger.debug("Retrying lost connection: %s", e)
                self.pool.close_all_connections()
                self.conn = None
                self.retry = True
                return True

        # Else, just clean conn and let exception bubble.
        self.pool.putconn(self.conn)
        self.conn = None


class ConnectionPool(ThreadedConnectionPool):
    def __init__(self, app, observers=None, **kw):
        self.app = app
        self.observers = observers
        super(ConnectionPool, self).__init__(**kw)

    def _connect(self, *a, **kw):
        return retry_connect(
            super(ConnectionPool, self)._connect, self.app, *a, **kw)

    def retry_connection(self):
        # Helper to retry a code block on connection lost.
        #
        # Manage pooled connection lost. Yield a context manager for one or two
        # attempt. The first attempt uses the connection as returned by the
        # pool. The second attempt closes pool and request a fresh connection.
        #
        # for attempt in postgres.retry_connection_pool():
        #     with attempt() as conn:
        #         conn.queryscalar("SELECT 1")
        #
        manager = RetryManager(self)
        for try_ in 0, 1:
            yield manager
            if not manager.retry:
                break

    def close_all_connections(self):
        # Close all connection, keeping pool opened.
        for conn in self._pool + list(self._used.values()):
            conn.close()
            self.putconn(conn)


def retry_connect(connect, app, *a, **kw):
    # Retry 3 times, 1 second wait between each try. This allows to wait for
    # Postgres to restart, but give up to consider Postgres as unavailable.
    for wait in [1] * 3 + [0]:
        try:
            conn = connect(*a, **kw)
            if not app.status.data['postgres']['available']:
                logger.info("Recovered Postgres connexion.")
                app.postgres.notify_observers()
            app.status.data['postgres']['available'] = True
            conn.set_session(autocommit=True)
            return conn
        except Exception as e:
            app.status.data['postgres']['available'] = False
            if wait:
                logger.debug("Retrying connection open in %ss: %s", wait, e)
                sleep(wait)
            else:  # Last wait is 0. Just give up.
                raise



import functools
import inspect
import logging
import json
from datetime import timedelta

from bottle import (
    Bottle,
    HTTPError, HTTPResponse, Response,
    default_app, request, response,
)

from ..toolkit.utils import JSONEncoder, utcnow
from ..toolkit.http import format_date
from ..toolkit.signing import (
    InvalidSignature,
    canonicalize_request,
    verify_v1,
)


logger = logging.getLogger(__name__)


def create_app(temboard):
    Response.default_content_type = 'text/plain'
    app = CustomBottle(autojson=False)
    app.temboard = temboard
    app.add_hook('before_request', before_request_log)
    app.add_hook('after_request', after_request_log)
    # First declared, first executed.
    app.install(JSONPlugin())
    app.install(ErrorPlugin())
    app.install(SignaturePlugin())
    app.install(PostgresPlugin())
    return app


class CustomBottle(Bottle):
    def default_error_handler(self, res):
        # Skip HTML template.
        return res.body

    def mount(self, prefix, app, **options):
        for plugin in self.plugins:
            app.install(plugin)
        return super(CustomBottle, self).mount(prefix, app, **options)


def before_request_log():
    logger.debug("New web request: %s %s", request.method, request.path)


def after_request_log():
    if response.status_code > 500:
        logmethod = logger.error
    else:
        logmethod = logger.info
    logmethod("%s %s %s", request.method, request.path, response.status)


class PostgresPlugin(object):
    def __init__(self):
        self._pool = None
        self._dbpool = None

    @property
    def dbpool(self):
        if not self._dbpool:
            self._dbpool = default_app().temboard.postgres.dbpool()
        return self._dbpool

    @property
    def pool(self):
        if not self._pool:
            self._pool = default_app().temboard.postgres.pool()
        return self._pool

    def apply(self, callback, route):
        wanted = self.wants_postgres(callback)
        if not wanted:
            return callback

        @functools.wraps(callback)
        def wrapper(*a, **kw):
            if 'pgpool' in wanted:
                # dbpool is not threadsafe! But wsgiref simple server is
                # not threaded.
                kw['pgpool'] = self.dbpool

            # Assume callbacks idempotence.
            for attempt in self.pool.retry_connection():
                with attempt() as conn:
                    if 'pgconn' in wanted:
                        kw['pgconn'] = conn

                    return callback(*a, **kw)

                if attempt.retry:
                    # Propagate connection lost to dbpool.
                    self.dbpool.closeall()

        return wrapper

    def wants_postgres(self, callback):
        argspec = inspect.getargspec(callback)
        return [a for a in argspec.args if a in ('pgconn', 'pgpool')]


class ErrorPlugin(object):
    def apply(self, callback, route):
        @functools.wraps(callback)
        def wrapper(*a, **kw):
            try:
                response = callback(*a, **kw)
            except HTTPError as e:
                if isinstance(e.body, str):
                    e.body = {'error': e.body}
                # Use HTTPResponse to customize body.
                response = HTTPResponse(e.body, e.status)
            except HTTPResponse as e:
                response = e
            except Exception:
                logger.exception("Unhandled error:")
                response = HTTPResponse({'error': 'Internal error.'}, 500)

            return response
        return wrapper


class JSONPlugin(object):
    # Bottle jsonify only dict. JSON Array was a security issue for some
    # browser.
    def apply(self, callback, route):
        @functools.wraps(callback)
        def wrapper(*a, **kw):
            res = callback(*a, **kw)
            body = res.body if isinstance(res, HTTPResponse) else res

            if not isinstance(body, (dict, list)):
                return res

            body = json.dumps(body, cls=JSONEncoder)
            if not isinstance(res, HTTPResponse):
                res = response.copy(cls=HTTPResponse)
            res.body = body

            res.headers['Content-Type'] = 'application/json'

            return res
        return wrapper


class SignaturePlugin(object):
    name = 'signature'

    def setup(self, app):
        self.app = app

    def apply(self, callback, route):
        @functools.wraps(callback)
        def wrapper(*a, **kw):
            request.username = self.authenticate()
            return callback(*a, **kw)
        return wrapper

    def authenticate(self):
        app = default_app().temboard

        date = request.headers['x-temboard-date']
        oldest_date = format_date(utcnow() - timedelta(hours=2))
        if date < oldest_date:
            raise HTTPError(400, "Request older than 2 hours.")

        signature = request.headers['x-temboard-signature']
        version, _, signature = signature.partition(':')
        if 'v1' != version:
            raise HTTPError(400, 'Unsupported signature format')

        if not signature:
            raise HTTPError(400, 'Malformed signature')

        path = request.environ['RAW_PATH_INFO']
        if request.environ['QUERY_STRING']:
            path = path + '?' + request.environ['QUERY_STRING']
        canonical_request = canonicalize_request(
            request.method, path, request.headers, request.body.read(),
        )

        try:
            verify_v1(app.config.signing_key, signature, canonical_request)
        except InvalidSignature:
            raise HTTPError(403, 'Invalid signature')

        user = request.headers.get('x-temboard-user')
        if not user:
            raise HTTPError(400, 'Missing username')

        return user



import os.path
from os import listdir
import datetime

from bottle import HTTPError


class ConfigurationFileManager:

    @classmethod
    def get_file_content(self, filepath, version=None):
        return read_file_content(filepath, version)

    @classmethod
    def save_file_content(self, filepath, filecontent, new_version=False):
        return save_file_content(filecontent, filepath, new_version)

    @classmethod
    def remove_version(self, filepath, version):
        return delete_file_version(filepath, version)

    @classmethod
    def get_versions(self, filepath):
        return get_file_versions(filepath)


def check_version_format(version):
    """
    Checks if a version number is well formed, eg:
        YYYY-MM-DDTHH:mm:ss
    """
    try:
        datetime.datetime.strptime(version, "%Y-%m-%dT%H:%M:%S")
        return True
    except Exception:
        raise HTTPError(406,
                        "Bad version format, should be 'YYYY-MM-DDTHH:mm:ss'")


def save_file_content(content, filepath, new_version=False):
    ret = {'last_version': None}
    if new_version is True and os.path.isfile(filepath):
        """
        When new_version param. is true, we need to save current file as a new
        version before writing new file content.
        """
        # Build new version's file path.
        dt_str = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        filepath_version = "{}.{}".format(filepath, dt_str)
        ret['last_version'] = dt_str
        # Read current version's content.
        cur_content = None
        with open(filepath) as fd:
            cur_content = fd.read()
        # Check if new version's file exists.
        if os.path.isfile(filepath_version):
            raise HTTPError(500, "Unable to create a new version, file %s "
                            "already exists." % (filepath_version))
        # Write current version's content in new version's file.
        with open(filepath_version, 'w') as fd:
            fd.write(cur_content)
    else:
        versions = get_file_versions(filepath)
        if len(versions) > 0:
            ret['last_version'] = versions[-1]

    # Write new file content into current file.
    with open(filepath, 'w') as fd:
        fd.write(content)

    ret['filepath'] = filepath
    return ret


def read_file_content(filepath, version=None):
    """
    Returns file content or content from a previsous version.
    """
    if version is not None:
        # Check the version is well formatted.
        check_version_format(version)
        filepath_version = "{}.{}".format(filepath, version)
        if not os.path.isfile(filepath_version):
            raise HTTPError(404, "Version %s of file %s does not exist."
                                 % (version, filepath))
        filepath = filepath_version

    with open(filepath) as fd:
        return fd.read()


def get_file_versions(filepath):
    """
    Returns a list of version number for a file path.
    """
    filedir = os.path.dirname(filepath)
    filename = os.path.basename(filepath)
    if not os.path.isdir(filedir):
        raise HTTPError(500, "Unable to list content from directory: %s"
                             % (filedir))
    if not os.path.isfile(filepath):
        raise HTTPError(404, "File %s does not exist." % (filepath))

    versions = []
    l_filename = len(filename)
    for f in listdir(filedir):
        try:
            if f[:l_filename] == filename \
               and check_version_format(f[l_filename + 1:]):
                """
                Let's consider f as one of the  previous version of the file
                if the first part of f's name is equal to original file name
                and the rest of f's name (-1 for the seperating dot) is a
                valid version number.
                """
                versions.append(f[l_filename + 1:])
        except Exception:
            pass
    # Return a sorted versions list.
    return sorted(versions, reverse=True)


def delete_file_version(filepath, version):
    """
    Remove a previous version of file.
    """
    check_version_format(version)
    if version in get_file_versions(filepath):
        filepath_version = "{}.{}".format(filepath, version)
        os.remove(filepath_version)
        return {'version': version, 'deleted': True}
    else:
        raise HTTPError(404, "Version %s of file %s does not exist."
                             % (version, filepath))



from setuptools import setup, find_packages, __version__ as setuptoolsv
import os

# Load version number
__version__ = None
setup_path = os.path.dirname(os.path.realpath(__file__))
exec(open(os.path.join(setup_path, 'temboardagent', 'version.py')).read())

if setuptoolsv < '1.0':
    __version__ = __version__.replace('+', '.')

setup(**dict(
    name='temboard-agent',
    version=__version__,
    description='PostgreSQL Remote Control Agent',
    author='Dalibo',
    author_email='contact@dalibo.com',
    license='PostgreSQL',
    url='https://labs.dalibo.com/temboard',
    classifiers=[
        "Environment :: No Input/Output (Daemon)",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: System :: Monitoring",
    ],
    python_requires='~=3.5',
    include_package_data=True,
    install_requires=[
        'bottle >= 0.12, <0.13',
        'cryptography',
    ],
    data_files=[
        ('share/temboard-agent/', [
            'share/auto_configure.sh',
            'share/purge.sh',
            'share/restart-all.sh',
        ]),
        ('share/temboard-agent/quickstart/', [
            'share/temboard-agent_CHANGEME.pem',
            'share/temboard-agent_CHANGEME.key',
            'share/temboard-agent_ca_certs_CHANGEME.pem',
        ]),
        ('lib/systemd/system', ['packaging/temboard-agent@.service']),
    ],
    entry_points={
        'console_scripts': [
            'temboard-agent = temboardagent.__main__:main',
        ],
        'temboardagent.plugins': [
            'activity = temboardagent.plugins.activity:ActivityPlugin',
            'administration = temboardagent.plugins.administration:AdministrationPlugin',  # noqa
            'dashboard = temboardagent.plugins.dashboard:DashboardPlugin',
            'maintenance = temboardagent.plugins.maintenance:MaintenancePlugin',  # noqa
            'monitoring = temboardagent.plugins.monitoring:MonitoringPlugin',
            'pgconf = temboardagent.plugins.pgconf:PgConfPlugin',
            'statements = temboardagent.plugins.statements:StatementsPlugin',
        ],
    },
    packages=find_packages(),
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
))



import sys
import os
import atexit

# Some global vars we need to have because signal handler function can't take
# extra parameters.
PIDFILE = None


def remove_pidfile(pidfile):
    """
    Delete the pidfile.
    """
    if pidfile:
        try:
            os.remove(pidfile)
        except OSError:
            pass


def daemonize(pidfile):
    """
    Run temboard-agent as a background daemon.
    Inspired by:
    http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
    """
    # Try to read pidfile
    try:
        with open(pidfile) as pf:
            pid = int(pf.read().strip())
    except (OSError, ValueError):
        pid = None

    # If pidfile exists, yet another process is probably running.
    if pid:
        sys.stderr.write("FATAL: pidfile %s already exist.\n" % pidfile)
        sys.exit(1)

    # Try to write pidfile.
    try:
        with open(pidfile, 'w+') as pf:
            pf.write("\0")
    except OSError:
        sys.stderr.write("FATAL: can't write pidfile %s.\n" % pidfile)
        sys.exit(1)

    # First fork.
    try:
        pid = os.fork()
        if pid > 0:
            # Exit first parent.
            sys.exit(0)
    except OSError as e:
        sys.stderr.write("FATAL: fork failed: %d (%s)\n"
                         % (e.errno, e.strerror))
        sys.exit(1)

    # Decouple from parent environment.
    os.chdir("/")
    os.setsid()
    os.umask(0)

    # Do second fork.
    try:
        pid = os.fork()
        if pid > 0:
            # Exit from second parent.
            sys.exit(0)
    except OSError as e:
        sys.stderr.write("FATAL: fork failed: %d (%s)\n"
                         % (e.errno, e.strerror))
        sys.exit(1)

    # Redirect standard file descriptors.
    sys.stdout.flush()
    sys.stderr.flush()
    si = open('/dev/null')
    so = open('/dev/null', 'a+')
    se = open('/dev/null', 'a+')
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())
    # write pidfile
    atexit.register(remove_pidfile, pidfile)
    pid = str(os.getpid())
    with open(pidfile, 'w+') as pf:
        pf.write("%s\n" % pid)
    global PIDFILE
    PIDFILE = pidfile



from base64 import b64decode, b64encode
from hashlib import sha256

from cryptography.hazmat.backends.openssl import backend as openssl_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.exceptions import InvalidSignature

from .errors import TemboardError
from .pycompat import quote_plus
from .utils import ensure_bytes


__all__ = [
    'InvalidSignature',
]


_REQUIRED_HEADERS = {
    'host',
    'x-temboard-date',
    'x-temboard-request-id',
    'x-temboard-user',
}

_REQUIRED_POST_HEADERS = _REQUIRED_HEADERS | {
    'content-type',
    'content-length',
}


def canonicalize_request(method, path, headers, body=b''):
    method = ensure_bytes(method.upper())
    if b'POST' == method:
        required = _REQUIRED_POST_HEADERS
    else:
        required = _REQUIRED_HEADERS

    canonicalized_headers = {
        name.lower(): value
        for name, value in
        headers.items()
        if name.lower() in required
    }

    missing_headers = required - set(canonicalized_headers)
    if missing_headers:
        missing_headers = ", ".join(sorted(missing_headers))
        raise TemboardError(
            "Missing headers for request signing: %s" % missing_headers)

    lines = [
        b"%s %s" % (method, ensure_bytes(quote_plus(path, safe='/'))),
        b"",
    ] + sorted([
        b"%s: %s" % (ensure_bytes(name), ensure_bytes(value))
        for name, value in
        canonicalized_headers.items()
    ])

    if b'POST' == method:
        h = sha256()
        h.update(body)
        lines += [
            b"",
            h.hexdigest().encode('ascii'),
        ]

    lines.append(b"")   # Final EOF

    return b"\n".join(lines)


def load_private_key(data):
    return serialization.load_pem_private_key(
        data, password=None, backend=openssl_backend)


def load_public_key(data):
    return serialization.load_pem_public_key(
        data, backend=openssl_backend)


HASH = hashes.SHA256()
PADDING = padding.PSS(
    mgf=padding.MGF1(HASH),
    salt_length=padding.PSS.MAX_LENGTH,
)


def sign_v1(private_key, payload):
    signature = private_key.sign(payload, PADDING, HASH)
    signature64 = b64encode(signature)
    return signature64.decode('ascii')


def verify_v1(public_key, signature, payload):
    signature_bin = b64decode(signature)
    public_key.verify(signature_bin, payload, PADDING, HASH)



# Note that our home-made background task implementation, we use a different
# semantic that state of the art background task implemtation:
#
# a task function is called a worker
# a worker process is called workerpool
# a message is called a task
# the message broker is called taskmanager
#

import logging.config

from ..toolkit.app import SubCommand
from ..toolkit.taskmanager import RunTaskMixin
from .app import app


logger = logging.getLogger(__name__)


@app.command
class RunTask(RunTaskMixin, SubCommand):
    """ Run a task foreground. """

    def main(self, args):
        workers = self.iter_workers()

        if '?' == args.worker_name:
            self.print_workers(workers)
        else:
            worker, worker_args = self.compute_worker_args(workers, args)
            worker(*worker_args)
        return 0
